/*
 * net.c
 *
 * Copyright 2004, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 * ͥåȥ̥롼
 */


#include<types.h>
#include<errno.h>
#include<lib.h>
#include<mm.h>
#include<proc.h>
#include<lock.h>
#include<fs.h>
#include<net/arp.h>
#include<net/icmp.h>
#include<net/udp.h>
#include<net/tcp.h>
#include<net/net.h>
#include<share/socket.h>


/************************************************************************************
 *
 * Misc
 *
 ************************************************************************************/


typedef struct RECV_BUF{
	struct RECV_BUF *next;
}RECV_BUF;


/*
 * GLOBAL
 * åη׻
 * parameters : data1 address,data2 address,byte size1,byte size2
 * return : פ
 */
ushort calcSum(uint *data,int size)
{
	union{
		unsigned long long u64;
		uint               u32[2];
		ushort             u16[4];
	}sum;
	uint tmp;


	sum.u64=0;
	for(;size>=sizeof(uint);size-=sizeof(uint))
		sum.u64+=*data++;
	if(size>0)sum.u64+=*data&((1<<(size*8))-1);

	tmp=sum.u32[1];
	sum.u32[1]=0;
	sum.u64+=tmp;
	tmp=sum.u32[1];
	sum.u32[1]=0;
	sum.u32[0]+=tmp;

	tmp=sum.u16[1];
	sum.u16[1]=0;
	sum.u32[0]+=tmp;
	tmp=sum.u16[1];
	sum.u16[1]=0;
	sum.u16[0]+=tmp;

	return ~sum.u16[0];
}

/*
 * Note!
 *  size1uintǳڤˤƤ롣
 */
ushort calcSumM(uint *data1,uint *data2,int size1,int size2)
{
	union{
		unsigned long long u64;
		uint               u32[2];
		ushort             u16[4];
	}sum;
	uint tmp;


	sum.u64=0;
	for(;size1>=sizeof(uint);size1-=sizeof(uint))
		sum.u64+=*data1++;
	for(;size2>=sizeof(uint);size2-=sizeof(uint))
		sum.u64+=*data2++;
	if(size2>0)sum.u64+=*data2&((1<<(size2*8))-1);

	tmp=sum.u32[1];
	sum.u32[1]=0;
	sum.u64+=tmp;
	tmp=sum.u32[1];
	sum.u32[1]=0;
	sum.u32[0]+=tmp;

	tmp=sum.u16[1];
	sum.u16[1]=0;
	sum.u32[0]+=tmp;
	tmp=sum.u16[1];
	sum.u16[1]=0;
	sum.u16[0]+=tmp;

	return ~sum.u16[0];
}


/*
 * GLOBAL
 * Close socket.
 * parameters : socket address
 * return : 0 or error number
 */
void closeSocket(SOCKET *sc)
{
	/* åȤγ */
	if(sc->stat&SOCK_CNCT_ON)
		releaseSckFromConect(sc);
	else kfree(sc);
}


/************************************************************************************
 *
 * Systemcall interface
 *
 ************************************************************************************/


enum{
	IP_PROTNUM_MAX=17,		/* IP ̥ץȥʥС͡ */
	SOCKET_MAX=10,			/* ץλĥåȤκ */
};


static SOCKET_INFO *sockInfo[IP_PROTNUM_MAX+1];


int sys_socket(int domain, int type, int protocol)
{
	int num;
	SOCKET *sc;


	switch(domain)
	{
		case AF_INET:
			switch(type)
			{
				case SOCK_DGRAM:
					if((protocol==0)||(protocol==IPPROTO_UDP))num=IPPROTO_UDP;
					else return -EPROTONOSUPPORT;
					break;
				case SOCK_STREAM:
					if((protocol==0)||(protocol==IPPROTO_TCP))num=IPPROTO_TCP;
					else return -EPROTONOSUPPORT;
					break;
				case SOCK_RAW:
					if((protocol==0)||(protocol==IPPROTO_ICMP))num=IPPROTO_ICMP;
					else return -EPROTONOSUPPORT;
					break;
				default:
					return -EPROTONOSUPPORT;
			}
			break;
		default:
			return -EPROTONOSUPPORT;
	}

	if((sc=kmalloc(sizeof(SOCKET)))==NULL)return -ENOMEM;
	sc->form=FD_FORM_NETSOKET;
	sc->sinf=sockInfo[num];
	sc->buf=NULL;
	sc->update=0;
	sc->wait.flag=0;
	sc->stat=0;
	sc->next=sc->prev=sc;
	sc->waitCnt=0;
	sc->refCount=1;

	/* ץϿ */
	return setToFd(sc);
}


int sys_send(int sd,const void *msg,size_t len,int flags,struct sockaddr *to)
{
	SOCKET *sc;


	if((sc=(SOCKET*)getFd(sd))==NULL)return -EBADF;
	if(checkMem(to)==-1)return -EFAULT;
	((SOCKET_INFO*)sc->sinf)->send(sc,msg,len,flags,to);

	return len;
}


int sys_recv(int sd,void *buf,size_t len,int flags,struct sockaddr *from,socklen_t *fromlen)
{
	SOCKET *sc;


	if((sc=(SOCKET*)getFd(sd))==NULL)return -EBADF;
	if(checkMem(buf)==-1)return -EFAULT;

	return ((SOCKET_INFO*)sc->sinf)->recv(sc,buf,len,flags,from,fromlen);
}


int sys_connect(int sockfd, const struct sockaddr *serv_addr)
{
	SOCKET *sc;


	if((sc=(SOCKET*)getFd(sockfd))==NULL)return -EBADF;
	if(checkMem((void*)serv_addr)==-1)return -EFAULT;

	return ((SOCKET_INFO*)sc->sinf)->connect(sc,serv_addr);
}


int sys_shutdown(int s, int how)
{
	SOCKET *sc;


	if((sc=(SOCKET*)getFd(s))==NULL)return -EBADF;

	return ((SOCKET_INFO*)sc->sinf)->shutdown(sc,how);
}


/************************************************************************************
 *
 * 
 *
 ************************************************************************************/


/*
 * GLOBAL
 * Regist socket information.
 * parameters : SOCKET_INFO address,protocol number
 * return : 0
 */
int registSocket(SOCKET_INFO *sinf,int prot)
{
	sockInfo[prot]=sinf;

	return 0;
}


/*
 * GLOBAL
 * Init network.
 */
int initNet()
{
	initArp();
	initIcmp();
	initUdp();
	initTcp();

	return 0;
}
