/*
 * tcp.c
 *
 * Copyright 2004, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 * Transmission control Protocl.
 */


#include<types.h>
#include<lib.h>
#include<errno.h>
#include<mm.h>
#include<proc.h>
#include<time.h>
#include<lock.h>
#include<interrupt.h>
#include<net/net.h>
#include<net/ether.h>
#include<net/ip.h>
#include<net/tcp.h>
#include<test.h>


static void closeConnect(SOCKET*,TCP_HEADER*);


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


enum{
	/* TCP flag. */
	TCP_FLG_FIN=1<<0,
	TCP_FLG_SYN=1<<1,
	TCP_FLG_RST=1<<2,
	TCP_FLG_PSH=1<<3,
	TCP_FLG_ACK=1<<4,
	TCP_FLG_URG=1<<5,

	TCP_WIN_SIZE=0xffff,	/* Window size. */
	TCP_MSS_SIZE=1460,
};


/*
 * PUBLIC
 * إåꡣ
 * parameters : dectination IP,pseudo header address,data size
 */
static inline void setPseudoHead(in_addr_t dstip,in_addr_t srcip,uint16_t size,PSEUDO_HEADER *pshead)
{
	pshead->dstip=dstip;
	pshead->srcip=srcip;
	pshead->tmp=0;
	pshead->prot=IPPROTO_TCP;
	pshead->len=swapWord(size);
}


/**************************************************************************
 *
 * إåץ
 *
 **************************************************************************/


enum{
	/* Header option. */
	TCP_KIND_END=  0,
	TCP_KIND_NOP=  1,
	TCP_KIND_MSS=  2,
	TCP_KIND_WSC=  3,
	TCP_KIND_TSTMP=8,

	TCP_OPT_MAXSIZE=32,		/* TCP ץκ祵 */
};


typedef struct{
	uint32_t timeStamp;
	uint32_t timeEcho;
	uint16_t mss;
	uint8_t  winscal;
}OPTIONS;


/*
 * PUBLIC
 * Get TCP options.
 * parameters : option address
 */
static void getTcpOptions(uchar *c,int size,OPTIONS *optst)
{
	uchar *last;


	last=c+size;
	while(c<last)
	{
		switch(*c)
		{
			case TCP_KIND_MSS:
				optst->mss= *(uint16_t*)(c+2);
				c+=4;
				break;
			case TCP_KIND_WSC:
				optst->winscal=*(uint8_t*)(c+2);
				c+=3;
				break;
			case TCP_KIND_TSTMP:
				optst->timeStamp=*(uint32_t*)(c+2);
				optst->timeEcho=*(uint32_t*)(c+6);
				c+=10;
				break;
			case TCP_KIND_NOP:
				c+=1;
				break;
			case TCP_KIND_END:
				break;
			default:
				c+=*(c+1);
		}
	}
}


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


/*
 * PUBLIC
 * Transmit a tcp segment.
 * Note!
 *  ץǡɬintǳڤ륵ˤ뤳ȡ
 * parameters : socket,TCP flag,option address,option size,data address,data size,fragment flag
 */
static void transSegment(SOCKET *sc,uint8_t flag,void *opt,int opt_size,void *data,int data_size,int fragment)
{
	char head[sizeof(PSEUDO_HEADER)+sizeof(TCP_HEADER)+TCP_OPT_MAXSIZE];
	TRANS_BUF_INFO tbi;
	TCP_HEADER *send_tcp;


	send_tcp=(TCP_HEADER*)(head+sizeof(PSEUDO_HEADER));

	/* إåꡣ */
	setPseudoHead(sc->dstip,getSrcip(sc->dstip),sizeof(TCP_HEADER)+opt_size,(PSEUDO_HEADER*)head);

	/* Хåեơ֥ */
	tbi.data[2]=data;
	tbi.size[2]=data_size;
	tbi.data[1]=(char*)send_tcp;
	tbi.size[1]=sizeof(TCP_HEADER)+opt_size;
	tbi.ipType=IPPROTO_TCP;

	/* Make TCP header. */
	send_tcp->srcport= sc->srcport;
	send_tcp->dstport= sc->dstport;
	send_tcp->seqnum=  swapInt32(sc->seqnum);
	send_tcp->acknum=  swapInt32(sc->acknum);
	send_tcp->headlen= ((sizeof(TCP_HEADER)+opt_size)/sizeof(uint32_t))<<4;
	send_tcp->flag=    flag;
	send_tcp->wndsize= swapWord(TCP_WIN_SIZE);
	send_tcp->chksum=  0;
	send_tcp->urgpoint=0;
	memcpy(send_tcp->option,opt,opt_size);
	send_tcp->chksum=calcSumM((uint*)head,data,sizeof(PSEUDO_HEADER)+sizeof(TCP_HEADER)+opt_size,data_size);
/************************************************************************************************************************
printk("send    %x,%x,%x,%x,%x\n",
	swapWord(send_tcp->srcport),swapWord(send_tcp->dstport),swapInt32(send_tcp->seqnum),swapInt32(send_tcp->acknum),
	send_tcp->flag);
************************************************************************************************************************/
	/*  */
	transIp(&tbi,sc->dstip,0,fragment);
}


/*
 * PUBLIC
 * ͥ󤬤ʤΥꥻåȥȤ
 * parameters : receive IP header
 */
static inline void transReset(IP_HEADER *ip)
{
	SOCKET sc;
	TCP_HEADER *tcp;


	tcp=(TCP_HEADER*)ip->data;
	sc.seqnum= 0;
	sc.acknum= swapInt32(tcp->seqnum)+1;
	sc.srcport=tcp->dstport;
	sc.dstport=tcp->srcport;
	sc.dstip=  ip->srcip;
	transSegment(&sc,TCP_FLG_RST,NULL,0,NULL,0,1);
}


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


enum{
	WATI_COUNT_MAX=3,		/* åȥХåեԤ祫ȿ */
};


/* PUBLIC */
static SOCKET cnctSocket;	/* ͥ󥯡 */
static int cnctLock;		/* ͥѥåȡ */


/*
 * PRIVATE
 * Save receive data.
 * parameters : IP data
 * return : 0 or task switch=1
 */
static int saveRecv(IP_HEADER *ip)
{
	TCP_HEADER *tcp;
	SOCKET *sc;


	tcp=(TCP_HEADER*)ip->data;
/**************************************************************************************************************
printk("receive %x,%x,%x,%x,%x\n",
	swapWord(tcp->srcport),swapWord(tcp->dstport),swapInt32(tcp->acknum),swapInt32(tcp->seqnum),tcp->flag);
**************************************************************************************************************/
	/* ͥ󥽥åȤθ */
	enter_spinlock(&cnctLock);
	{
		for(sc=cnctSocket.next;;sc=sc->next)
		{
			if(sc==&cnctSocket)
			{
				exit_spinlock(&cnctLock);
				transReset(ip);
				return 0;
			}

			if((sc->srcport==tcp->dstport)&&(sc->dstport==tcp->srcport)&&(sc->dstip==ip->srcip))
				break;
		}

		while(sc->update==1);
		sc->update=1;								/* åȹϡ */
	}
	exit_spinlock(&cnctLock);

	if(sc->stat&SOCK_CNCT_TRANS_FIN)
		closeConnect(sc,tcp);
	else
	{
		/*
		 * ǡ򥽥åȤϤ
		 * ХåեƤʤȤϺåԤġ
		 */
		for(;;)
		{
			if(sc->buf==NULL)
			{
				sc->buf=ip;
				sc->waitCnt=0;
				break;
			}
			if(sc->waitCnt>=WATI_COUNT_MAX)break;
			wait_task();
			sc->waitCnt+=1;
		}

	}

	sc->update=0;									/* åȹλ */

	/* READԤץε */
	if(sc->wait.flag!=0)
	{
		wake_intr(&sc->wait);
		return 1;
	}

	return 0;
}


/*
 * PUBLIC
 * Get data from receive buffer.
 * parameters : socket,segment address buffer
 * return : receive segment size or error number
 */
static int waitReceive(SOCKET *sc,TCP_HEADER **tcp_buf)
{
	enum{REPLY_TIMEOUT=5000};	/* ॢȥߥá */

	IP_HEADER *ip;


	for(;;)
	{
		/* 쥷֥Хåեˤꡣ */
		if(sc->buf!=NULL)break;

		/* ǡ쥷Ԥ */
		sc->wait.flag=0;
		wait_intr(&sc->wait,REPLY_TIMEOUT);

		/*  */
		if(sc->wait.flag)		/* Time out. */
		{
			sc->wait.flag=0;
			if(sc->buf==NULL)return -ETIMEDOUT;
			break;
		}
	}

	ip=(IP_HEADER*)sc->buf;
	sc->buf=NULL;
	*tcp_buf=(TCP_HEADER*)ip->data;

	return swapWord(ip->len)-(ip->verhead&0xf)*4;
}


/*
 * PUBLIC
 * 㤦Ȥ줿Ȥν
 * parameters : socket,receive TCP header
 * return : error number
 */
static int recvBadSegment(SOCKET *sc,TCP_HEADER *tcp)
{
	if(tcp->flag&TCP_FLG_RST)return -ECONNREFUSED;

	/* ꥻåȥȤ */
	sc->acknum=swapInt32(tcp->seqnum)+1;
	sc->seqnum=0;
	transSegment(sc,TCP_FLG_RST,NULL,0,NULL,0,1);

	return -EAGAIN;
}


/*
 * GROBAL
 * Receive TCP datagram.
 * parameters : IP data
 * return : 0 or task switch=1
 */
int receiveTcp(IP_HEADER *ip)
{
	int tcp_size;
	PSEUDO_HEADER pshead;


	/* åγǧ */
	tcp_size=swapWord(ip->len)-(ip->verhead&0xf)*4;
	setPseudoHead(ip->dstip,ip->srcip,tcp_size,&pshead);
	if(calcSumM((uint*)&pshead,(uint*)ip->data,sizeof(PSEUDO_HEADER),tcp_size))
		return 0;

	return saveRecv(ip);
}


/**************************************************************************
 *
 * ͥ
 *
 **************************************************************************/


enum{
	CNCT_CLOSE_WAIT_TIME=5000,		/* ǽॢȻms */
};


typedef struct{
	uint8_t  kind;
	uint8_t  len;
	uint16_t mss;
}MSS;


/* PRIVATE */
static SOCKET *finSocket;		/* 楽åȥ󥯡 */


/*
 * PRIVATE
 * Add to connection link.
 * parameters : socket
 */
static inline void addCnctLink(SOCKET *sc)
{
	enter_spinlock(&cnctLock);
	{
		sc->next=cnctSocket.next;
		sc->prev=&cnctSocket;
		cnctSocket.next->prev=sc;
		cnctSocket.next=sc;
	}
	exit_spinlock(&cnctLock);
}


/*
 * PRIVATE
 * Delete from connection link.
 * parameters : socket
 */
static inline void delCnctLink(SOCKET *sc)
{
	enter_spinlock(&cnctLock);
	{
		while(sc->update==1);
		sc->next->prev=sc->prev;
		sc->prev->next=sc->next;
		sc->next=sc->prev=sc;
	}
	exit_spinlock(&cnctLock);
}


/*
 * PRIVATE
 * ǳϥåȤ󥯤³롣
 * parameters : socket
 */
static inline void addCloseSocket(SOCKET *sc)
{
	sc->waitNext=finSocket;
	finSocket=sc;
	sc->waitTime=sys_time();
}


/*
 * PRIVATE
 * ǳϥåȤǡॢȤõ
 */
static inline void delCloseTimeout()
{
	uint time;
	SOCKET *p;


	time=sys_time();
	while(finSocket!=NULL)
	{
		if(time-finSocket->waitTime>=CNCT_CLOSE_WAIT_TIME)
		{
			p=finSocket->waitNext;
			delCnctLink(finSocket);
			kfree(finSocket);
			finSocket=p;
		}
		else break;
	}
}


/*
 * PUBLIC
 * Get sequence number.
 * return : sequence number
 */
static inline uint32_t getSeqNumber()
{
	/* 4msǣ䤹RFC793ˡ */
	return rdtsc()/clock_1m*4;
}


/*
 * PUBLIC
 * Get ephemeral port number.
 * return : ephemeral port number
 */
static ushort getEphemeralPort()
{
	enum{BEGIN_EPHM_PORT=1024};		/* եݡȤκǽֹ档 */

	static int gate=0;
	static int volatile port=BEGIN_EPHM_PORT;
	int pt;


	enter_spinlock(&gate);
	{
		pt=port++;
	}
	exit_spinlock(&gate);

	return pt;
}


/*
 * PUBLIC
 * Close connection.
 */
static void closeConnect(SOCKET *sc,TCP_HEADER *tcp)
{
	uint32_t ack;


	if(tcp->flag&TCP_FLG_FIN)												/* FINȤ */
	{
		ack=(sc->stat&SOCK_CNCT_RECV_FINACK)?sc->seqnum:sc->seqnum-1;
		if(swapInt32(tcp->acknum)==ack)
		{
			sc->stat|=SOCK_CNCT_RECV_FIN;
			sc->acknum+=1;
			transSegment(sc,TCP_FLG_ACK,NULL,0,NULL,0,1);
		}
	}
	else if((tcp->flag&TCP_FLG_ACK)&&(swapInt32(tcp->acknum)==sc->seqnum))	/* ACKȤ */
		sc->stat|=SOCK_CNCT_RECV_FINACK;

	sc->waitTime=sys_time();
}


/*
 * PUBLIC
 * transmit fin segment.
 * parameters :
 */
static inline void transFin(SOCKET *sc)
{
	/* FINȤ */
	transSegment(sc,TCP_FLG_FIN|TCP_FLG_ACK,NULL,0,NULL,0,1);
	sc->seqnum+=1;
	sc->stat|=SOCK_CNCT_TRANS_FIN;

	/* Хåե˻Ĥꤢꡣ */
	if(sc->buf!=NULL)closeConnect(sc,(TCP_HEADER*)sc->buf->data);
	sc->buf=NULL;

	sc->waitTime=sys_time();
	delCloseTimeout();				/* ॢȥåȤ롣 */
	addCloseSocket(sc);				/* åȥ󥯤³롣 */
}


/*
 * PUBLIC
 * ͥΩ
 * parameters : socket
 * return : 0 or error number
 */
static int connectTo(SOCKET *sc)
{
	int rest;
	int tcp_size;
	MSS mss;
	TCP_HEADER *recv_tcp;
	OPTIONS optst;


	/* ͥ󥯤³ */
	addCnctLink(sc);

	/* إåץꡣ */
	mss.kind=TCP_KIND_MSS;
	mss.len= 4;
	mss.mss= swapWord(TCP_MSS_SIZE);

	/* SYNȤ */
	transSegment(sc,TCP_FLG_SYN,&mss,sizeof(mss),NULL,0,1);
	sc->seqnum+=1;

	/* Ȥμ */
	if((rest=waitReceive(sc,&recv_tcp))<0)goto ERR;
	tcp_size=rest;
	if((swapInt32(recv_tcp->acknum)!=sc->seqnum)||((recv_tcp->flag&(TCP_FLG_SYN|TCP_FLG_ACK))!=(TCP_FLG_SYN|TCP_FLG_ACK)))
	{
		rest=recvBadSegment(sc,recv_tcp);
		goto ERR;
	}

	/* ץγǧ(MSSΤ߼) */
	optst.mss=TCP_MSS_SIZE;
	getTcpOptions(recv_tcp->option,tcp_size-sizeof(TCP_HEADER),&optst);
	sc->mss=optst.mss;

	/* ǧ */
	sc->acknum=swapInt32(recv_tcp->seqnum)+1;
	sc->stat=SOCK_CNCT_ON;
	transSegment(sc,TCP_FLG_ACK,NULL,0,NULL,0,1);

	return 0;
ERR:
	/* ͥ󥯤 */
	delCnctLink(sc);

	return rest;
}


/*
 * GLOBAL
 * åȤ򥳥ͥ󤫤鳫롣
 * parameters : socket
 */
void releaseSckFromConect(SOCKET *sc)
{
	if((sc->stat&SOCK_CNCT_TRANS_FIN)==0)
			transFin(sc);
}


/**************************************************************************
 *
 * Socket interface.
 *
 **************************************************************************/


static void send(SOCKET *sc,const void *msg,size_t len,int flags,struct sockaddr *to)
{

}


/*
 * return : copy data size or error number
 */
static int recv(SOCKET *sc,void *buf,size_t len,int flags,struct sockaddr *from,socklen_t *fromlen)
{
	char *data;
	int tcp_size,data_size,all_size;
	TCP_HEADER *tcp;
	struct RESERVE{
		TCP_HEADER *tcp;
		int         size;
	}reserve={NULL};


	/* γǧ */
	if((sc->stat&SOCK_CNCT_ON)==0) return -ENOTCONN;

	all_size=0;
	for(;;)
	{
		/*  */
		if((tcp_size=waitReceive(sc,&tcp))<0)return tcp_size;

		/* γǧ */
RET:	if(swapInt32(tcp->acknum)==sc->seqnum)
		{
			if(tcp->flag&TCP_FLG_RST)return -ECONNREFUSED;
		
			if(swapInt32(tcp->seqnum)==sc->acknum)	
			{
				/* λȤμ */
				if(tcp->flag&TCP_FLG_FIN)	
				{
					sc->stat|=SOCK_CNCT_RECV_FIN|SOCK_CNCT_RECV_FINACK;
					sc->acknum+=1;
					break;
				}
			}
			else if(swapInt32(tcp->seqnum)>sc->acknum)
			{
				/*
				 * ȤΥֹ椬礭ȤϡΥȤ夷ǽΤ
				 * αƤ
				 */
				reserve.tcp=tcp;
				reserve.size=tcp_size;
				continue;
			}
			else continue;
		}
		else return recvBadSegment(sc,tcp);	

		/* 桼Хåե˥ԡ */
		data_size=tcp_size-(tcp->headlen>>4)*4;
		data=(char*)tcp+(tcp_size-data_size);
		all_size+=data_size;
		if(all_size>len)
		{
			data_size-=all_size-len;
			all_size=len;
			memcpy(buf,data,data_size);
			break;
		}
		memcpy(buf,data,data_size);
		(char*)buf+=data_size;
	
		/* αȤ硣 */
		if(reserve.tcp!=NULL)
		{
			tcp=reserve.tcp;
			tcp_size=reserve.size;
			reserve.tcp=NULL;
			goto RET;
		}
	
		if(sc->buf==NULL)break;
	}

	/* Ȥ */
	sc->acknum+=all_size;
	transSegment(sc,TCP_FLG_ACK,NULL,0,NULL,0,1);

	return all_size;
}


static int connect(SOCKET *sc,const struct sockaddr *addr)
{
	if(sc->stat&SOCK_CNCT_ON)return -EISCONN;

	/* Set connect structure. */
	sc->next=sc->prev=sc;
	sc->dstip=((struct sockaddr_in*)addr)->sin_addr.s_addr;
	sc->dstport=((struct sockaddr_in*)addr)->sin_port;
	sc->srcport=swapWord(getEphemeralPort());
	sc->seqnum=getSeqNumber();
	sc->acknum=0;

	return connectTo(sc);
}


static int shutdown(SOCKET *sc,int flag)
{
	if((sc->stat&SOCK_CNCT_ON)==0)return -ENOTCONN;
	if((sc->stat&SOCK_CNCT_TRANS_FIN)==0)
		transFin(sc);

	return 0;
}


/**************************************************************************
 *
 * 
 *
 **************************************************************************/


static SOCKET_INFO sockInfo={send,recv,connect,shutdown};


/*
 * GLOBAL
 * Init UDP.
 */
int initTcp()
{
	registSocket(&sockInfo,IPPROTO_TCP);
	cnctSocket.next=cnctSocket.prev=&cnctSocket;

	return 0;
}

/***************************************************************************/
void test_tcp()
{
	printk("waitNext=%x\n",*(uint*)0x16be64);
}
/****************************************************************************/
